import copy
from typing import Dict

from blueness import module
from blue_options.logger import log_dict, log_list
from blue_objects import file, objects
from blue_objects.metadata import get_from_object

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.web.functions import url_to_filename
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


def expanding_the_extractions(
    script: BaseScript,
    node_name: str,
) -> bool:
    map_node_name = "extraction"

    crawl_cache: Dict[str, str] = get_from_object(
        script.object_name,
        "web_crawl_cache",
        {},
    )
    log_dict(logger, "using", crawl_cache, "crawled url(s)")

    list_of_urls = [
        url
        for url, content_type in crawl_cache.items()
        if "html" in content_type
        and not file.exists(
            objects.path_of(
                object_name=script.object_name,
                filename="{}_cache/{}.txt".format(
                    map_node_name,
                    url_to_filename(url),
                ),
            )
        )
    ]
    log_list(logger, "using", list_of_urls, "crawled unextracted html(s).")

    max_nodes = min(
        len(list_of_urls),
        script.nodes[node_name]["max_nodes"],
    )
    logger.info(
        "{}: expanding {} X {}...".format(
            NAME,
            map_node_name,
            max_nodes,
        )
    )

    map_node = script.nodes[map_node_name]
    del script.nodes[map_node_name]
    script.G.remove_node(map_node_name)

    reduce_node_name = "generating_summary"
    for index in range(max_nodes):
        url = list_of_urls[index]
        index_node_name = f"{map_node_name}_{index+1:03d}"

        success, url_content = file.load_yaml(
            filename=objects.path_of(
                object_name=script.object_name,
                filename="web_crawl_cache/{}.yaml".format(
                    url_to_filename(url),
                ),
            ),
        )
        if not success:
            logger.warning(f"{url}: failed to load url content.")
            continue
        if "text" not in url_content:
            logger.warning(f"{url}: no text found in url content.")
            continue

        logger.info(f"{url} -{map_node_name}-> {index_node_name}")

        script.nodes[index_node_name] = copy.deepcopy(map_node)

        script.nodes[index_node_name]["prompt"] = map_node["prompt"].replace(
            ":::url_content",
            url_content["text"],
        )

        script.nodes[index_node_name]["url"] = url
        script.nodes[index_node_name]["cache"] = "{}_cache/{}.txt".format(
            map_node_name,
            url_to_filename(url),
        )

        script.G.add_node(index_node_name)
        script.G.add_edge(
            index_node_name,
            node_name,
        )
        script.G.add_edge(
            reduce_node_name,
            index_node_name,
        )

    script.nodes_changed = True

    return script.save_graph()



from blue_objects import file, path

from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.repository.orbital_data_explorer.actions import (
    dict_of_actions,
)


class OrbitalDataExplorerScript(BaseScript):
    name = path.name(file.path(__file__))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dict_of_actions.update(dict_of_actions)



import copy

from blueness import module
from blue_objects import file, path

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.repository.blue_amo.actions import dict_of_actions
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


class BlueAmoScript(BaseScript):
    name = path.name(file.path(__file__))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dict_of_actions.update(dict_of_actions)

    def generate_graph(
        self,
        log: bool = True,
        verbose: bool = False,
        save_graph: bool = True,
    ) -> bool:
        if not super().generate_graph(
            log=log,
            verbose=verbose,
            save_graph=save_graph,
        ):
            return False

        map_node_name = "generating_the_frames"
        if log:
            logger.info(
                "{}: expanding {} X {}...".format(
                    NAME,
                    map_node_name,
                    self.vars["frame_count"],
                )
            )

        map_node = self.nodes[map_node_name]
        del self.nodes[map_node_name]
        self.G.remove_node(map_node_name)

        reduce_node_name = "stitching_the_frames"
        for index in range(self.vars["frame_count"]):
            node_name = f"generating_frame_{index+1:03d}"

            self.nodes[node_name] = copy.deepcopy(map_node)

            self.G.add_node(node_name)
            self.G.add_edge(
                node_name,
                "setting_frame_prompts",
            )
            self.G.add_edge(
                reduce_node_name,
                node_name,
            )

        return self.save_graph(log=log) if save_graph else True



from typing import List
from functools import reduce

from blue_options.terminal import show_usage, xtra

from blue_assistant.script.repository import list_of_script_names
from blue_assistant.script.repository.functions import get_script_versions


def help_list(
    tokens: List[str],
    mono: bool,
) -> str:
    args = [
        "[--delim +]",
        "[--log 0]",
    ]

    return show_usage(
        [
            "@assistant",
            "script",
            "list",
        ]
        + args,
        "list scripts.",
        mono=mono,
    )


def help_run(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("~download,dryrun,~upload", mono=mono)

    script_options = "script=<name>,version=<version>"

    args = [
        "[--test_mode 1]",
        "[--verbose 1]",
        "[--runnable <~node_1,~node_2>]",
    ]

    def script_version_details(script_name: str) -> List[str]:
        list_of_script_versions = get_script_versions(script_name)

        return (
            [
                "{}: {}".format(
                    script_name,
                    " | ".join(list_of_script_versions),
                )
            ]
            if list_of_script_versions
            else []
        )

    return show_usage(
        [
            "@assistant",
            "script",
            "run",
            f"[{options}]",
            f"[{script_options}]",
            "[-|<object-name>]",
        ]
        + args,
        "run <script-name>/<script-version> in <object-name>.",
        {
            "name: {}".format(" | ".join(list_of_script_names)): reduce(
                lambda x, y: x + y,
                [
                    script_version_details(script_name)
                    for script_name in list_of_script_names
                ],
                [],
            )
        },
        mono=mono,
    )


help_functions = {
    "list": help_list,
    "run": help_run,
}



from typing import List, Tuple

from blue_options import string

from blue_assistant.script.load import load_script


def get_script_versions(
    script_name: str,
) -> List[str]:
    success, script = load_script(
        script_name=script_name,
        object_name=string.timestamp(),
        log=False,
        save_graph=False,
    )

    return list(script.versions.keys()) if success else []



import os
from typing import Dict, List
from functools import reduce
import networkx as nx
from tqdm import tqdm

from blue_options.options import Options
from blue_objects import file, path, objects
from blue_objects.metadata import post_to_object, get_from_object
from blueflow.workflow import dot_file

from blue_assistant.logger import logger


class RootScript:
    name = path.name(file.path(__file__))

    def __init__(
        self,
        object_name: str,
        script_version: str = "base",
        test_mode: bool = False,
        log: bool = True,
        verbose: bool = False,
        save_graph: bool = True,
    ):
        self.valid = False

        self.nodes_changed = False

        self.object_name = object_name

        self.test_mode = test_mode

        self.verbose = verbose

        self.dict_of_actions = {}

        metadata_filename = os.path.join(
            file.path(__file__),
            f"../{self.name}",
            "metadata.yaml",
        )
        self.metadata: Dict
        success, self.metadata = file.load_yaml(metadata_filename)
        if not success:
            logger.error(f"cannot load {self.name}/metadata.yaml")
            return

        # validate and set
        last_script_version = get_from_object(
            object_name,
            "script_version",
            script_version,
        )
        if script_version != last_script_version:
            logger.error(
                "script version in {} is {} != {}".format(
                    object_name,
                    last_script_version,
                    script_version,
                )
            )
            return
        if not post_to_object(
            object_name,
            "script_version",
            script_version,
        ):
            return
        if log:
            logger.info(f"script version: {script_version}")

        self.metadata.setdefault("script", {})
        if not isinstance(self.script, dict):
            logger.error(
                "script: expected dict, received {}.".format(
                    self.script.__class__.__name__,
                )
            )
            return

        self.script.setdefault("nodes", {})
        if not isinstance(
            self.nodes,
            dict,
        ):
            logger.error(
                "nodes: expected dict, received {}.".format(
                    self.nodes.__class__.__name__,
                )
            )
            return

        self.script.setdefault("vars", {})
        if not isinstance(
            self.vars,
            dict,
        ):
            logger.error(
                "vars: expected dict, received {}.".format(
                    self.vars.__class__.__name__,
                )
            )
            return

        self.script.setdefault("versions", {})
        if not isinstance(
            self.versions,
            dict,
        ):
            logger.error(
                "versions: expected dict, received {}.".format(
                    self.versions.__class__.__name__,
                )
            )
            return

        if script_version != "base":
            if script_version not in self.versions:
                logger.error(f"{script_version}: script version not found.")
                return

            for thing in ["vars", "nodes"]:
                updates = self.versions[script_version].get(thing, {})
                if len(updates) and log:
                    logger.info(f"{thing}.update({updates})")
                self.metadata["script"][thing].update(updates)

        if self.test_mode:
            if log:
                logger.info("🧪  test mode is on.")

            if "test_mode" in self.script:
                updates = self.script["test_mode"]
                if log:
                    logger.info(f"🧪  vars.update({updates})")
                self.vars.update(updates)

            for node_name, node in self.nodes.items():
                if "test_mode" in node:
                    updates = node["test_mode"]
                    if log:
                        logger.info(f"🧪  {node_name}.update({updates})")
                    node.update(updates)

        if log:
            logger.info(
                "loaded {} node(s): {}".format(
                    len(self.nodes),
                    ", ".join(self.nodes.keys()),
                )
            )

            logger.info(
                "loaded {} var(s): {}".format(
                    len(self.vars),
                    ", ".join(self.vars.keys()),
                )
            )
        if verbose:
            for var_name, var_value in self.vars.items():
                logger.info("{}: {}".format(var_name, var_value))

        if not self.generate_graph(
            log=log,
            verbose=verbose,
            save_graph=save_graph,
        ):
            return

        self.valid = True

    def __str__(self) -> str:
        return "{}[{} var(s), {} node(s) -> {}]".format(
            self.__class__.__name__,
            len(self.vars),
            len(self.nodes),
            self.object_name,
        )

    def apply_vars(self, text: str) -> str:
        for var_name, var_value in self.vars.items():
            text = text.replace(f":::{var_name}", str(var_value))

        for node_name, node in self.nodes.items():
            node_output = node.get("output", "")
            if isinstance(node_output, str):
                text = text.replace(f":::{node_name}", node_output)

        return text

    def generate_graph(
        self,
        log: bool = True,
        verbose: bool = False,
        save_graph: bool = True,
    ) -> bool:
        self.G: nx.DiGraph = nx.DiGraph()

        list_of_nodes = list(self.nodes.keys())
        for node in self.nodes.values():
            list_of_nodes += node.get("depends-on", "").split(",")

        list_of_nodes = list({node_name for node_name in list_of_nodes if node_name})
        if verbose:
            logger.info(
                "{} node(s): {}".format(
                    len(list_of_nodes),
                    ", ".join(list_of_nodes),
                )
            )

        for node_name in list_of_nodes:
            self.G.add_node(node_name)

        for node_name, node in self.nodes.items():
            for dependency in node.get("depends-on", "").split(","):
                if dependency:
                    self.G.add_edge(node_name, dependency)

        return self.save_graph(log=log) if save_graph else True

    def get_context(
        self,
        node_name: str,
    ) -> List[str]:
        return reduce(
            lambda x, y: x + y,
            [self.get_context(successor) for successor in self.G.successors(node_name)],
            [node_name],
        )

    def perform_action(
        self,
        node_name: str,
    ) -> bool:
        action_name = self.nodes[node_name].get("action", "unknown")
        logger.info(f"---- node: {node_name} ---- ")

        if action_name in self.dict_of_actions:
            return self.dict_of_actions[action_name](
                script=self,
                node_name=node_name,
            )

        logger.error(f"{action_name}: action not found.")
        return False

    def run(
        self,
        runnable: str = "",
    ) -> bool:
        logger.info(f"{self.name}.run -> {self.object_name}")

        if runnable:
            logger.info(f"applying runnables: {runnable}")
            runnable_options = Options(runnable)
            for node_name, node_is_runnable in runnable_options.items():
                logger.info(f"{node_name}.runnable={node_is_runnable}")
                self.nodes[node_name]["runnable"] = node_is_runnable

        success: bool = True
        while (
            not all(self.nodes[node].get("completed", False) for node in self.nodes)
            and success
        ):
            self.nodes_changed = False

            for node_name in tqdm(self.nodes):
                if self.nodes[node_name].get("completed", False):
                    continue

                if not self.nodes[node_name].get("runnable", True):
                    logger.info(f"Not runnable, skipped: {node_name}.")
                    self.nodes[node_name]["completed"] = True
                    continue

                pending_dependencies = [
                    node_name_
                    for node_name_ in self.G.successors(node_name)
                    if not self.nodes[node_name_].get("completed", False)
                ]
                if pending_dependencies:
                    logger.info(
                        'node "{}": {} pending dependenci(es): {}'.format(
                            node_name,
                            len(pending_dependencies),
                            ", ".join(pending_dependencies),
                        )
                    )
                    continue

                if not self.perform_action(node_name=node_name):
                    success = False
                    break

                self.nodes[node_name]["completed"] = True

                cache_filename = self.nodes[node_name].get("cache", "")
                if cache_filename:
                    if not file.save_text(
                        objects.path_of(
                            object_name=self.object_name,
                            filename=cache_filename,
                        ),
                        [self.nodes[node_name].get("output", "")],
                    ):
                        success = False
                        break

                if self.nodes_changed:
                    logger.info("🪄  nodes changed.")
                    break

        if not post_to_object(
            self.object_name,
            "output",
            self.metadata,
        ):
            success = False

        return success

    def save_graph(
        self,
        log: bool = True,
    ) -> bool:
        return dot_file.save_to_file(
            objects.path_of(
                filename="workflow.dot",
                object_name=self.object_name,
            ),
            self.G,
            caption=" | ".join(
                [
                    self.name,
                    self.object_name,
                ]
            ),
            add_legend=False,
            log=log,
        )

    # Aliases
    @property
    def script(self) -> Dict:
        return self.metadata["script"]

    @property
    def nodes(self) -> Dict[str, Dict]:
        return self.metadata["script"]["nodes"]

    @property
    def vars(self) -> Dict:
        return self.metadata["script"]["vars"]

    @property
    def versions(self) -> Dict:
        return self.metadata["script"]["versions"]



import copy

from blueness import module
from blue_objects import file, path

from blue_assistant import NAME
from blue_assistant.script.repository.base.root import RootScript
from blue_assistant.script.actions import dict_of_actions
from blue_assistant.logger import logger


NAME = module.name(__file__, NAME)


class BaseScript(RootScript):
    name = path.name(file.path(__file__))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dict_of_actions = copy.deepcopy(dict_of_actions)



from typing import Dict, Callable

from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.repository.orbital_data_explorer.actions import (
    expanding_the_extractions,
)


dict_of_actions: Dict[str, Callable[[BaseScript, str], bool]] = {
    "expanding_the_extractions": expanding_the_extractions.expanding_the_extractions,
}



from typing import Dict, Callable

from blue_assistant.script.repository.base.root import RootScript
from blue_assistant.script.actions.generic import generic_action
from blue_assistant.script.actions.generate_image import generate_image
from blue_assistant.script.actions.generate_text import generate_text
from blue_assistant.script.actions.web_crawl import web_crawl


dict_of_actions: Dict[str, Callable[[RootScript, str], bool]] = {
    "generic": generic_action,
    "generate_image": generate_image,
    "generate_text": generate_text,
    "web_crawl": web_crawl,
}



from typing import List, Dict, Tuple

from blueness import module
from blue_options.logger import log_dict, log_list
from blue_objects import file
from blue_objects import objects
from blue_objects.metadata import get_from_object, post_to_object

from blue_assistant import NAME
from blue_assistant.web.fetch import fetch_links_and_text
from blue_assistant.web.functions import url_to_filename
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


def crawl_list_of_urls(
    seed_urls: List[str],
    object_name: str,
    max_iterations: int = 10,
    verbose: bool = False,
    cache_prefix: str = "",
) -> Tuple[bool, Dict[str, str]]:
    logger.info(
        "{}.crawl_list_of_urls({}): {} -> {}".format(
            NAME,
            len(seed_urls),
            ", ".join(seed_urls),
            object_name,
        )
    )

    queue: List[str] = list(
        set(
            seed_urls
            + get_from_object(
                object_name,
                f"{cache_prefix}_queue",
                [],
            )
        )
    )
    log_list(logger, "queue:", queue, "url(s)")

    crawl_cache: Dict[str, str] = get_from_object(
        object_name,
        f"{cache_prefix}_cache",
        {},
    )
    log_dict(logger, "loaded cache:", crawl_cache, "url(s)")

    iteration: int = 0
    while queue:
        url = queue[0]
        queue = queue[1:]

        logger.info(
            "{} [#{:,}/{:,}]: {} ".format(
                "✅ " if url in crawl_cache else "🔗 ",
                iteration,
                len(queue),
                url,
            )
        )
        if url in crawl_cache:
            continue

        url_summary = fetch_links_and_text(
            url=url,
            verbose=verbose,
        )
        content_type = url_summary.get("content_type", "unknown")

        if "html" in content_type:
            if not file.save_yaml(
                filename=objects.path_of(
                    object_name=object_name,
                    filename="{}_cache/{}.yaml".format(
                        cache_prefix,
                        url_to_filename(url),
                    ),
                ),
                data=url_summary,
            ):
                return False, {}

        crawl_cache[url] = content_type

        queue = (
            queue
            + url_summary.get("list_of_urls", [])
            + [
                url
                for url in url_summary.get("list_of_ignored_urls", [])
                if any(url.startswith(url_prefix) for url_prefix in seed_urls)
            ]
        )
        queue = list({url for url in queue if url not in crawl_cache.keys()})

        iteration += 1
        if max_iterations != -1 and iteration >= max_iterations:
            logger.warning(f"max iteration of {max_iterations} reached.")
            break

    if queue:
        logger.warning(f"queue: {len(queue)}")

    if not (
        post_to_object(
            object_name,
            f"{cache_prefix}_cache",
            crawl_cache,
        )
        and post_to_object(
            object_name,
            f"{cache_prefix}_queue",
            queue,
        )
    ):
        return False, {}

    log_dict(logger, "crawled", crawl_cache, "url(s)")
    log_list(logger, "queue:", queue, "url(s)")

    return True, crawl_cache
